"""
    Tools for aligning continuous and discrete events
    functions:
    get_time_array
    slice_inds_and_offsets
    index_of_nearest_value
    event_triggered_response
"""

import numpy as np
import pandas as pd


def get_time_array(
    t_start, t_end, sampling_rate=None, step_size=None, include_endpoint=True
):  # NOQA E501
    """
    A function to get a time array between two specified timepoints at a defined sampling rate  # NOQA E501
    Deals with possibility of time range not being evenly divisible by desired sampling rate  # NOQA E501
    Uses np.linspace instead of np.arange given decimal precision issues with np.arange (see np.arange documentation for details)  # NOQA E501

    Parameters:
    -----------
    t_start : float
        start time for array
    t_end : float
        end time for array
    sampling_rate : float
        desired sampling of array
        Note: user must specify either sampling_rate or step_size, not both
    step_size : float
        desired step size of array
        Note: user must specify either sampling_rate or step_size, not both
    include_endpoint : Boolean
        Passed to np.linspace to calculate relative time
        If True, stop is the last sample. Otherwise, it is not included.
            Default is True

    Returns:
    --------
    numpy.array
        an array of timepoints at the desired sampling rate

    Examples:
    ---------
    get a time array exclusive of the endpoint
    >>> t_array = get_time_array(
        t_start=-1,
        t_end=1,
        step_size=0.5,
        include_endpoint=False
    )

    np.array([-1., -0.5,  0.,  0.5])


    get a time array inclusive of the endpoint
    >>> t_array = get_time_array(
        t_start=-1,
        t_end=1,
        step_size=0.5,
        include_endpoint=False
    )

    np.array([-1., -0.5,  0.,  0.5, 1.0])


    get a time array where the range can't be evenly divided by the desired step_size
    in this case, the time array includes the last timepoint before the desired endpoint
    >>> t_array = get_time_array(
        t_start=-1,
        t_end=0.75,
        step_size=0.5,
        include_endpoint=False
    )

    np.array([-1., -0.5,  0.,  0.5])


    Instead of passing the step_size, we can pass the sampling rate
    >>> t_array = get_time_array(
        t_start=-1,
        t_end=1,
        sampling_rate=2,
        include_endpoint=False
    )

    np.array([-1., -0.5,  0.,  0.5])
    """
    assert (
        sampling_rate is not None or step_size is not None
    ), "must specify either sampling_rate or step_size"  # NOQA E501
    assert (
        sampling_rate is None or step_size is None
    ), "cannot specify both sampling_rate and step_size"  # NOQA E501

    # value as a linearly spaced time array
    if not step_size:
        step_size = 1 / sampling_rate
    # define a time array
    n_steps = (t_end - t_start) / step_size
    if n_steps != int(n_steps):
        # if the number of steps isn't an int, that means it isn't possible
        # to end on the desired t_after using the defined sampling rate
        # we need to round down and include the endpoint
        n_steps = int(n_steps)
        t_end_adjusted = t_start + n_steps * step_size
        include_endpoint = True
    else:
        t_end_adjusted = t_end

    if include_endpoint:
        # add an extra step if including endpoint
        n_steps += 1

    t_array = np.linspace(t_start, t_end_adjusted, int(n_steps), endpoint=include_endpoint)

    return t_array


def slice_inds_and_offsets(
    data_timestamps,
    event_timestamps,
    time_window,
    sampling_rate=None,
    include_endpoint=False,
):  # NOQA E501
    """
    Get nearest indices to event timestamps, plus ind offsets (start:stop)
    for slicing out a window around the event from the trace.
    Parameters:
    -----------
    data_timestamps : np.array
        Timestamps of the datatrace.
    event_timestamps : np.array
        Timestamps of events around which to slice windows.
    time_window : list
        [start_offset, end_offset] in seconds
    sampling_rate : float, optional, default=None
        Sampling rate of the datatrace.
        If left as None, samplng rate is inferred from data_timestamps.

    Returns:
    --------
    event_indices : np.array
        Indices of events from the timestamps provided.
    start_ind_offset : int
    end_ind_offset : int
    trace_timebase :  np.array
    """
    if sampling_rate is None:
        sampling_rate = 1 / np.diff(data_timestamps).mean()

    event_indices = index_of_nearest_value(data_timestamps, event_timestamps)
    trace_len = (time_window[1] - time_window[0]) * sampling_rate
    start_ind_offset = int(time_window[0] * sampling_rate)
    end_ind_offset = int(start_ind_offset + trace_len) + int(include_endpoint)
    trace_timebase = np.arange(start_ind_offset, end_ind_offset) / sampling_rate

    return event_indices, start_ind_offset, end_ind_offset, trace_timebase


def index_of_nearest_value(data_timestamps, event_timestamps):
    """
    The index of the nearest sample time for each event time.

    Parameters:
    -----------
    sample_timestamps : np.ndarray of floats
        sorted 1-d vector of data sample timestamps.
    event_timestamps : np.ndarray of floats
        1-d vector of event timestamps.

    Returns:
    --------
    event_aligned_ind : np.ndarray of int
        An array of nearest sample time index for each event times.
    """
    insertion_ind = np.searchsorted(data_timestamps, event_timestamps)
    # is the value closer to data at insertion_ind or insertion_ind-1?
    ind_diff = data_timestamps[insertion_ind] - event_timestamps
    ind_minus_one_diff = np.abs(
        data_timestamps[np.clip(insertion_ind - 1, 0, np.inf).astype(int)] - event_timestamps
    )

    event_indices = insertion_ind - (ind_diff > ind_minus_one_diff).astype(int)
    return event_indices


def event_triggered_response(  # noqa C901
    data,
    t,
    y,
    event_times,
    t_start=None,
    t_end=None,
    t_before=None,
    t_after=None,
    output_sampling_rate=None,
    include_endpoint=True,
    output_format="tidy",
    interpolate=True,
    censor=True,
    censor_times=None,
    nan_policy="error",
):  # NOQA E501
    """
    Slices a timeseries relative to a given set of event times
    to build an event-triggered response. From mindscope_utilities (commit af2b70a)

    For example, If we have data such as a measurement of neural activity
    over time and specific events in time that we want to align
    the neural activity to, this function will extract segments of the neural
    timeseries in a specified time window around each event.

    The times of the events need not align with the measured
    times of the neural data.
    Relative times will be calculated by linear interpolation.

    Parameters:
    -----------
    data: Pandas.DataFrame
        Input dataframe in tidy format
        Each row should be one observation
        Must contains columns representing `t` and `y` (see below)
    t : string
        Name of column in data to use as time data
    y : string
        Name of column to use as y data
    event_times: list or array of floats
        Times of events of interest.
        Values in column specified by `y` will be sliced and interpolated
            relative to these times
    t_start : float
        start time relative to each event for desired time window
        e.g.:   t_start = -1 would start the window 1 second before each
                t_start = 1 would start the window 1 second after each event
        Note: cannot pass both t_start and t_before
    t_before : float
        time before each of event of interest to include in each slice
        e.g.:   t_before = 1 would start the window 1 second before each event
                t_before = -1 would start the window 1 second after each event
        Note: cannot pass both t_start and t_before
    t_end : float
        end time relative to each event for desired time window
        e.g.:   t_end = 1 would end the window 1 second after each event
                t_end = -1 would end the window 1 second before each event
        Note: cannot pass both t_end and t_after
    t_after : float
        time after each event of interest to include in each slice
        e.g.:   t_after = 1 would start the window 1 second after each event
                t_after = -1 would start the window 1 second before each event
        Note: cannot pass both t_end and t_after
    output_sampling_rate : float
        Desired sampling of output.
        Input data will be interpolated to this sampling rate if interpolate = True (default). # NOQA E501
        If passing interpolate = False, the sampling rate of the input timeseries will # NOQA E501
        be used and output_sampling_rate should not be specified.
    include_endpoint : Boolean
        Passed to np.linspace to calculate relative time
        If True, stop is the last sample. Otherwise, it is not included.
            Default is True
    output_format : string
        'wide' or 'tidy' (default = 'tidy')
        if 'tidy'
            One column representing time
            One column representing event_number
            One column representing event_time
            One row per observation (# rows = len(time) x len(event_times))
        if 'wide', output format will be:
            time as indices
            One row per interpolated timepoint
            One column per event,
                with column names titled event_{EVENT NUMBER}_t={EVENT TIME}
    interpolate : Boolean
        if True (default), interpolates each response onto a common timebase
        if False, shifts each response to align indices to a common timebase
    censor: Boolean
        if True (default), censor observations that take place after the next event time
        if False, do not censor
    censor_times: list or array or None
        if None, and censor is True, then use event_times as the censor times
        if times are provided, then these are the times at which ETR is censored
    nan_policy: How to handle NaNs in the input data
        "error": raise an exception if NaNs are present in the time window of an ETR
        "interpolate": interpolate over NaN values
        "exclude": exclude any response with a NaN in the response window

    Returns:
    --------
    Pandas.DataFrame
        See description in `output_format` section above

    Examples:
    ---------
    An example use case, recover a sinousoid from noise:

    First, define a time vector
    >>> t = np.arange(-10,110,0.001)

    Now build a dataframe with one column for time,
    and another column that is a noise-corrupted sinuosoid with period of 1
    >>> data = pd.DataFrame({
            'time': t,
            'noisy_sinusoid': np.sin(2*np.pi*t) + np.random.randn(len(t))*3
        })

    Now use the event_triggered_response function to get a tidy
    dataframe of the signal around every event

    Events will simply be generated as every 1 second interval
    starting at 0, since our period here is 1
    >>> etr = event_triggered_response(
            data,
            x = 'time',
            y = 'noisy_sinusoid',
            event_times = np.arange(100),
            t_start = -1,
            t_end = 1,
            output_sampling_rate = 100
        )
    Then use seaborn to view the result
    We're able to recover the sinusoid through averaging
    >>> import matplotlib.pyplot as plt
    >>> import seaborn as sns
    >>> fig, ax = plt.subplots()
    >>> sns.lineplot(
            data = etr,
            x='time',
            y='noisy_sinusoid',
            ax=ax
        )
    """
    # ensure that non-conflicting time values are passed
    assert (
        t_before is not None or t_start is not None
    ), "must pass either t_start or t_before"  # noqa: E501
    assert (
        t_after is not None or t_end is not None
    ), "must pass either t_start or t_before"  # noqa: E501

    assert (
        t_before is None or t_start is None
    ), "cannot pass both t_start and t_before"  # noqa: E501
    assert t_after is None or t_end is None, "cannot pass both t_after and t_end"  # noqa: E501

    if interpolate is False:
        assert (
            output_sampling_rate is None
        ), "if interpolation = False, the sampling rate of the input timeseries will be used. Do not specify output_sampling_rate"  # NOQA E501

    # assign time values to t_start and t_end
    if t_start is None:
        t_start = -1 * t_before
    if t_end is None:
        t_end = t_after

    # ensure that t_end is greater than t_start
    assert t_end > t_start, "must define t_end to be greater than t_start"

    assert (not censor) or (output_format == "tidy"), "cannot censor data in wide output"

    assert nan_policy in ["error", "interpolate", "exclude"], "unrecognized nan_policy"

    if censor:
        event_times = np.sort(event_times)

    if output_sampling_rate is None:
        # if sampling rate is None,
        # set it to be the mean sampling rate of the input data
        output_sampling_rate = 1 / np.diff(data[t]).mean()

    # sort data by time column to prevent errors in interpolate
    data = data.sort_values(by=t)

    # if interpolate is set to True,
    # we will calculate a common timebase and
    # interpolate every response onto that timebase
    if interpolate:
        # set up a dictionary with key 'time' and
        t_array = get_time_array(
            t_start=t_start,
            t_end=t_end,
            sampling_rate=output_sampling_rate,
            include_endpoint=include_endpoint,
        )
        data_dict = {"time": t_array}

        # Add one extra timestep in the data outside the t_window
        # this ensures we are not extrapolating at the edge of t_window
        # and instead doing interpolation.
        dt = np.diff(data[t].values).mean()

        # iterate over all event times
        data_reindexed = data.set_index(t, inplace=False)

        for event_number, event_time in enumerate(np.array(event_times)):
            # get a slice of the input data surrounding each event time
            data_slice = data_reindexed[y].loc[
                event_time + t_start - dt : event_time + t_end + dt
            ]  # noqa: E501

            # if the slice is empty, we will fill it with NaNs
            if len(data_slice) == 0:
                data_dict.update(
                    {
                        "event_{}_t={}".format(event_number, event_time): np.full(
                            len(t_array), np.nan
                        )
                    }
                )

            elif np.any(np.isnan(data_slice)):
                if nan_policy == "error":
                    # raise exception
                    raise Exception("NaN value in data slice, at event time {}".format(event_time))
                elif nan_policy == "exclude":
                    # exclude this event
                    data_dict.update(
                        {
                            "event_{}_t={}".format(event_number, event_time): np.full(
                                len(t_array), np.nan
                            )
                        }
                    )
                else:
                    # Interpolate over NaNs
                    x_data = data_slice[~data_slice.isnull()]
                    data_slice[:] = np.interp(
                        data_slice.index.values, x_data.index.values, x_data.values
                    )

                    interpolated = np.interp(
                        data_dict["time"],
                        data_slice.index - event_time,
                        data_slice.values,
                    )
                    # Screen for extrapolation points
                    interpolated[data_dict["time"] < np.min(data_slice.index - event_time)] = np.nan
                    interpolated[data_dict["time"] > np.max(data_slice.index - event_time)] = np.nan
                    data_dict.update(
                        {"event_{}_t={}".format(event_number, event_time): interpolated}
                    )
            else:
                # update our dictionary to have a new key defined as
                # 'event_{EVENT NUMBER}_t={EVENT TIME}' and
                # a value that includes an array that represents the
                # sliced data around the current event, interpolated
                # on the linearly spaced time array
                interpolated = np.interp(
                    data_dict["time"],
                    data_slice.index - event_time,
                    data_slice.values,
                )
                # Screen for extrapolation points
                interpolated[data_dict["time"] < np.min(data_slice.index - event_time)] = np.nan
                interpolated[data_dict["time"] > np.max(data_slice.index - event_time)] = np.nan
                data_dict.update({"event_{}_t={}".format(event_number, event_time): interpolated})

        # define a wide dataframe as a dataframe of the above compiled dictionary  # NOQA E501
        wide_etr = pd.DataFrame(data_dict)

    # if interpolate is False,
    # we will calculate a common timebase and
    # shift every response onto that timebase
    else:
        (
            event_indices,
            start_ind_offset,
            end_ind_offset,
            trace_timebase,
        ) = slice_inds_and_offsets(  # NOQA E501
            np.array(data[t]),
            np.array(event_times),
            time_window=[t_start, t_end],
            sampling_rate=None,
            include_endpoint=True,
        )
        all_inds = event_indices + np.arange(start_ind_offset, end_ind_offset)[:, None]
        wide_etr = (
            pd.DataFrame(
                data[y].values.T[all_inds],
                index=trace_timebase,
                columns=[
                    "event_{}_t={}".format(event_index, event_time)
                    for event_index, event_time in enumerate(event_times)
                ],  # NOQA E501
            )
            .rename_axis(index="time")
            .reset_index()
        )

    if output_format == "wide":
        # return the wide dataframe if output_format is 'wide'
        return wide_etr.set_index("time")
    elif output_format == "tidy":
        # if output format is 'tidy',
        # transform the wide dataframe to tidy format
        # first, melt the dataframe with the 'id_vars' column as "time"
        tidy_etr = wide_etr.melt(id_vars="time")

        # add an "event_number" column that contains the event number
        tidy_etr["event_number"] = (
            tidy_etr["variable"].map(lambda s: s.split("event_")[1].split("_")[0]).astype(int)
        )

        # add an "event_time" column that contains the event time ()
        tidy_etr["event_time"] = tidy_etr["variable"].map(lambda s: s.split("t=")[1]).astype(float)

        # drop the "variable" column, rename the "value" column
        tidy_etr = tidy_etr.drop(columns=["variable"]).rename(columns={"value": y})
        # return the tidy event triggered responses
    if censor:
        tidy_etr = censor_event_triggered_response(
            tidy_etr, y, t_start, t_end, event_times, censor_times
        )
    return tidy_etr


def censor_event_triggered_response(etr, y, t_start, t_end, event_times, censor_times=None):
    """
    censors the event triggered response by the immediately preceeding or
    subsequent event times if that event time is within the (t_start, t_end)
    time window

    censored timepoints are replaced with NaN, so all data points are still present

    etr: dataframe, event triggered response
    y: column of the response variable to censor
    t_start: start of event triggered response window
    t_end: end of event triggered response window
    censor: Boolean
        if True, censor observations that take place after the next event time
        if False, do not censor
    censor_times: list or array or None
        if None, and censor is True, then use event_times as the censor times
        if times are provided, then these are the times at which ETR is censored
    """

    if censor_times is None:
        # Compute when we should censor
        diff = np.diff(event_times)
        diff_backward = np.concatenate([[np.inf], diff])
        diff_forward = np.concatenate([diff, [np.inf]])
        backward_time = [-np.min([np.abs(t_start), x]) for x in diff_backward]
        forward_time = [np.min([t_end, x]) for x in diff_forward]

        # double check we have all events
        assert len(event_times) == len(etr["event_number"].unique()), "event times missing"

        # Censor trials
        for index, time in enumerate(event_times):
            vec = (etr["event_number"] == index) & (etr["time"] < backward_time[index])
            etr.loc[vec, y] = np.nan
            vec = (etr["event_number"] == index) & (etr["time"] > forward_time[index])
            etr.loc[vec, y] = np.nan

        return etr
    else:
        censor_times = np.sort(censor_times)
        backward_time = []
        forward_time = []
        for e in event_times:
            before = censor_times[censor_times < e]
            if len(before) == 0:
                backward_time.append(t_start)
            else:
                backward_time.append(np.max([t_start, before[-1] - e]))
            after = censor_times[censor_times > e]
            if len(after) == 0:
                forward_time.append(t_end)
            else:
                forward_time.append(np.min([t_end, after[0] - e]))

        # Censor trials
        for index, time in enumerate(event_times):
            vec = (etr["event_number"] == index) & (etr["time"] < backward_time[index])
            etr.loc[vec, y] = np.nan
            vec = (etr["event_number"] == index) & (etr["time"] > forward_time[index])
            etr.loc[vec, y] = np.nan

        return etr



"""
Important utility functions for formatting the data
    get_subject_assets
    generate_data_asset_attach_params
    attach_data
    check_data_assets
    add_data_asset_path
    get_all_df_for_nwb
    get_foraging_model_info
"""

import os
import warnings

import numpy as np
import pandas as pd
from aind_analysis_arch_result_access.han_pipeline import get_mle_model_fitting
from aind_data_access_api.document_db import MetadataDbClient
from codeocean import CodeOcean
from codeocean.data_asset import DataAssetAttachParams

from aind_dynamic_foraging_data_utils import nwb_utils


def get_subject_assets(subject_id, **kwargs):
    """
    Returns the docDB results for a subject. If duplicate entries exist, take the last
    based on processing time. Skips pavlovian task.

    equivalent to get_assets(subjects=[subject_id])

    subject_id (str or int) subject id to get assets for from docDB
    processed (bool) if True, look for processed assets. If False, look for raw assets
    task (list of strings), if empty, include all task variants: Uncoupled Baiting,
        Coupled Baiting, Uncoupled Without Baiting, Coupled Without Baiting.
        If not empty, only include the task variants provided.
    modality (list of strings), required data modality. If empty list, does not filter
        modalities should the data modality abbreviations, for example: behavior,
        behavior-videos, fib, ecephys
    stage (list of strings), if empty, include all training stages. Otherwise, only
        return stages included in this list. Possible stage names include STAGE_1,
        STAGE_1_WARMUP, STAGE_2, STAGE_3, STAGE_4, STAGE_FINAL, GRADUATED, None
    extra_filter (dict), docdb query

    Example
    results = get_subject_assets(my_id)
    co_assets = attach_data(results['code_ocean_asset_id'].values)

    """
    return get_assets(subjects=[subject_id], **kwargs)


def get_assets(  # NOQA: C901
    subjects=[],
    processed=True,
    task=[],
    modality=["behavior"],
    stage=[],
    extra_filter={},
    input_projection={},
):
    """
    Returns the docDB results for a subject. If duplicate entries exist, take the last
    based on processing time. Skips pavlovian task.

    subjects (a list of strs or ints) subject ids to get assets for from docDB
    processed (bool) if True, look for processed assets. If False, look for raw assets
    task (list of strings), if empty, include all task variants: Uncoupled Baiting,
        Coupled Baiting, Uncoupled Without Baiting, Coupled Without Baiting.
        If not empty, only include the task variants provided.
    modality (list of strings), required data modality. If empty list, does not filter
        modalities should the data modality abbreviations, for example: behavior,
        behavior-videos, fib, ecephys
    stage (list of strings), if empty, include all training stages. Otherwise, only
        return stages included in this list. Possible stage names include STAGE_1,
        STAGE_1_WARMUP, STAGE_2, STAGE_3, STAGE_4, STAGE_FINAL, GRADUATED, None
    extra_filter (dict), docdb query

    Example
    results = get_assets(subjects=[my_id])
    co_assets = attach_data(results['code_ocean_asset_id'].values)
    """
    # Create metadata client
    client = MetadataDbClient(
        host="api.allenneuraldynamics.org", database="metadata_index", collection="data_assets"
    )

    # Filter by task
    if len(task) > 0:
        task_filter = {"$or": []}
        for t in task:
            task_filter["$or"].append(
                {
                    "session.session_type": {"$regex": "^{}".format(t)},
                }
            )
    else:
        task_filter = {
            "session.session_type": {"$regex": "^(Uncoupled|Coupled)( Without)? Baiting"},
        }

    # Filter by data modality
    if len(modality) > 0:
        modality_filter = {"$and": []}
        for m in modality:
            modality_filter["$and"].append(
                {"data_description.modality.abbreviation": {"$regex": m}}
            )
    else:
        modality_filter = {}

    # Filter by stage:
    if len(stage) > 0:
        stage_filter = {
            "session.stimulus_epochs.output_parameters.task_parameters.stage_in_use": {"$in": stage}
        }
    else:
        stage_filter = {}

    # Do we want processed or raw assets
    if processed:
        processed_string = "_.*processed_[0-9-_]*"
    else:
        processed_string = (
            "_[0-9][0-9][0-9][0-9]-[0-9][0-9]-[0-9][0-9]_[0-9][0-9]-[0-9][0-9]-[0-9][0-9]"
        )

    # Query based on subject id
    if len(subjects) == 0:
        print("Query will be slow without explicit subject ids")
        subject_filter = {
            "name": {"$regex": "^behavior_[0-9]*{}$".format(processed_string)},
        }
        # Return only essential information for performance
        projection = {
            "name": 1,
            "_id": 1,
            "session": 1,
            "session_name": 1,
            "external_links": 1,
            "subject.subject_id": 1,
            **input_projection,
        }
    else:
        subject_filter = {
            "name": {
                "$regex": "^behavior_("
                + "|".join([str(x) for x in subjects])
                + "){}$".format(processed_string)
            },
        }
        # Return all information
        projection = None

    # Query
    results = pd.DataFrame(
        client.retrieve_docdb_records(
            filter_query={
                **subject_filter,
                **task_filter,
                **modality_filter,
                **stage_filter,
                **extra_filter,
            },
            projection=projection,
        )
    )

    # If nothing is found, return
    if len(results) == 0:
        print("No results found for {}".format(subjects))
        return

    # look for duplicate entries, taking the last by processing time
    results["session_name"] = [x.split("_processed")[0] for x in results["name"]]
    results = results.sort_values(by="name")
    results_no_duplicates = results.drop_duplicates(subset="session_name", keep="last").copy()

    # If there were duplicates, make a warning and print the duplicates
    if len(results) != len(results_no_duplicates):
        duplicated = results[results.duplicated(subset="session_name", keep=False)]
        warnings.warn("Duplicate session entries in docDB")
        for index, row in duplicated.iterrows():
            print("duplicated: {}".format(row["name"]))

    # Make code ocean ID a column
    results_no_duplicates["code_ocean_asset_id"] = [
        link["Code Ocean"][0] if "Code Ocean" in link else ""
        for link in results_no_duplicates["external_links"]
    ]

    return results_no_duplicates.reset_index(drop=True)


def generate_data_asset_attach_params(data_asset_IDs, mount_point=None):
    """
    generate_data_asset_attach_params is a helper function for attach_data
    data_asset_IDs:  list of data asset IDs, i.e. the 16 hash string for the data asset in CO.
    mount_point: the mount point (folder) for the data asset. Default is None.
    """
    data_assets = []
    for ID in data_asset_IDs:
        if mount_point:
            data_assets.append(DataAssetAttachParams(id=ID, mount=mount_point))
        else:
            data_assets.append(DataAssetAttachParams(id=ID))
    return data_assets


def attach_data(data_asset_IDs, token_name="CUSTOM_KEY"):
    """
    attach_data attaches a list of data_asset_ID to the capsule.
    data_asset_IDs: list of data asset IDs, i.e. the 16 hash string for the data asset in CO.
    token_name: the name of the token in the environment variable. Default is CUSTOM_KEY.
                see more info here:
                https://docs.codeocean.com/user-guide/code-ocean-api/authentication#to-create-an-access-token

    Note that the list of data_asset_IDs should be <100 or you may risk CO crashing.
    Example
    results = get_subject_assets(my_id)
    co_assets = attach_data(results['code_ocean_asset_id'].values)
    """

    # Check for too many assets
    if len(data_asset_IDs) > 100:
        warnings.warn(
            "Attaching more than 100 data assets at one time may crash Code Ocean. "
            + "Please batch data into 100 sessions per call"
        )
        return

    # Get asset attach params
    data_assets = generate_data_asset_attach_params(data_asset_IDs, mount_point=None)

    # Configure api
    token = os.getenv(token_name)
    if not token:
        warnings.warn("no token found. please create token")
        return
    client = CodeOcean(domain="https://codeocean.allenneuraldynamics.org", token=token)
    capsule_id = os.getenv("CO_CAPSULE_ID")

    # Try to attach all assets
    try:
        results = client.capsules.attach_data_assets(
            capsule_id=capsule_id,
            attach_params=data_assets,
        )
        return results
    except Exception as e:
        print(e)
        print("Attaching assets in a batch failed, trying individually")

    # Try to attach assets one by one (slow)
    for asset in data_assets:
        try:
            results = client.capsules.attach_data_assets(
                capsule_id=capsule_id,
                attach_params=[asset],
            )
        except Exception as e:
            print("Could not attach this asset: {}".format(e.data[0]))
    return results


def check_data_assets(co_assets, data_asset_IDs):
    """
    co_assets: a list of DataAssetAttachResults, produced by attach_data()
    data_asset_IDs: a list of code ocean Data Assets, passed into attach_data()
    This function is delicate because CO is strange about "ready",
    but its a useful quick check
    """
    if all([x.ready for x in co_assets if x.id in data_asset_IDs]):
        print("all data assets are ready")
    else:
        print("some data assets are not ready")


def add_data_asset_path(results):
    """
    Adds the filepath to the dataframe
    Example
    results = get_subject_assets(my_id)
    results = add_data_asset_path(results)
    """
    results["data_path"] = [
        os.path.abspath(os.path.join(os.sep, "data", x["name"], "nwb", x["session_name"] + ".nwb"))
        for index, x in results.iterrows()
    ]
    return results


def get_all_df_for_nwb(filename_sessions, interested_channels=None):
    """
    get_all_df_for_nwb gets all the dataframes for the NWB and
    returns the df_trials, df_events, df_fip

    Returns:
        df_trials (pd.DataFrame): dataframe with each row a trial from sessions
        df_events (pd.DataFrame): dataframe with each row an event from sessions
        df_fip (pd.DataFrame): dataframe with each row a timepoint for a signal from sessions
    """
    print(f"Saving channels: {interested_channels}")

    # Lists to collect session-level DataFrames
    list_df_trials = []
    list_df_events = []
    list_df_fip = []

    for idx, nwb_file in enumerate(filename_sessions):
        nwb = nwb_utils.load_nwb_from_filename(nwb_file)
        ses_idx = nwb_file.split("/")[-1].replace("behavior_", "").rsplit("_", 1)[0]
        print(f"CURRENTLY RUNNING {idx+1}/{len(filename_sessions)}: {ses_idx}")
        print("--------------------------------------------------")

        # Try to process df_trials, df_fip, df_events, otherwise go to next nwb_file

        # Trials
        try:
            df_ses_trials = nwb_utils.create_df_trials(nwb)
            list_df_trials.append(df_ses_trials)
        except Exception as e:
            print(f"Skipping {ses_idx} due to error in df_trials: {e}")
            continue

        # FIP
        try:
            df_ses_fip = nwb_utils.create_df_fip(nwb, tidy=True)
            if interested_channels:
                df_ses_fip = df_ses_fip[df_ses_fip["event"].isin(interested_channels)]
            list_df_fip.append(df_ses_fip)
        except Exception as e:
            print(f"Skipping {ses_idx} due to error in df_fip: {e}")
            continue

        # Events
        try:
            df_ses_events = nwb_utils.create_df_events(nwb)
            df_ses_events["ses_idx"] = ses_idx  # Add session identifier
            list_df_events.append(df_ses_events)
        except Exception as e:
            print(f"Skipping {ses_idx} due to error in df_events: {e}")
            continue

    # Concatenate all collected DataFrames
    df_trials = pd.concat(list_df_trials, axis=0).reset_index(drop=True)
    df_events = pd.concat(list_df_events, axis=0).reset_index(drop=True)
    df_fip = pd.concat(list_df_fip, axis=0).reset_index(drop=True)

    return (df_trials, df_events, df_fip)


def get_foraging_model_info(
    df_trials, df_sess, loc=None, model_name="QLearning_L2F1_CKfull_softmax"
):
    """
    get_foraging_model_info: retrieves fitted foraging_model information
    df_trials: dataframe for trials (1 row per trials) from nwb_utils.create df_trials.
               saved df_trials_fm will have L_prob, R_prob, Q_left, Q_right

               (if choice kernel in model), L_kernel, R_kernel
    df_sess: dataframe for sessions (1 row per session) from nwb_utils.create_df_sessions
               saved df_sess_fm will have parameters fitted for each mouse.
    loc: location to save the updated df_trials, df_sess with suffix `_fm.csv`.
                If given, we will save, otherwise we will return the dataframes
    model_name: model alias to get the model information

    """
    print(f"Retrieving foraging model {model_name}")
    df_trials_fm = df_trials.copy()
    df_trials_fm = df_trials_fm.rename(columns={"animal_response": "choice"})
    df_trials_fm["choice_name"] = df_trials_fm["choice"].map({1: "right", 0: "left"})

    df_trials_fm["model_name"] = model_name
    df_trials_fm["L_prob"] = np.nan
    df_trials_fm["R_prob"] = np.nan
    df_trials_fm["Q_left"] = np.nan
    df_trials_fm["Q_right"] = np.nan

    # check if CK is in model_name, if so, add df_trials_fm['L_kernel]
    if "CK" in model_name:
        df_trials_fm["L_kernel"] = np.nan
        df_trials_fm["R_kernel"] = np.nan

    df_sess_params = []
    for index, sess_i in df_sess.iterrows():
        df = get_mle_model_fitting(
            subject_id=str(sess_i["subject_id"]),
            session_date=str(sess_i["session_date"]),
            agent_alias=model_name,
        )
        sess_idx = str(sess_i["ses_idx"])

        # edge case of two sessions in one day
        if len(df) > 1:
            num_trials_for_nwb = sess_i.finished_trials
            df = df.query(f"n_trials == {num_trials_for_nwb}")

        if (
            df is None
            or df.get("params") is None
            or df.get("latent_variables") is None
            or df["latent_variables"][0] is None
        ):
            print(
                f"Skipping {sess_idx}. Fitted model {model_name}, \
                 params, or latent variables not found for this session"
            )
            continue  # skip if no model fits is found for this session

        # Fitted parameters
        params = df["params"][0]
        params["ses_idx"] = str(sess_i["ses_idx"])
        df_sess_params.append(params)

        # Fitted latent variables
        fitted_latent = df["latent_variables"][0]

        # pull the information
        mouse_choice_idx = df_trials_fm.index[
            (df_trials_fm["ses_idx"] == sess_idx) & (df_trials_fm["choice"] < 2)
        ]

        # Add check on mouse_choice_idx and length of choice_prob

        qvals = np.array(fitted_latent["q_value"]).astype(float)
        choice_kernel = np.array(fitted_latent["choice_kernel"]).astype(float)
        choice_prob = np.array(fitted_latent["choice_prob"]).astype(float)

        if len(mouse_choice_idx) != np.shape(choice_prob)[1]:
            print(
                f"Skipping {sess_idx}. Fitted model {model_name} \
                  does not have matching number of trials"
            )
            continue  # skip if the fitted choices do not match number of trials

        df_trials_fm.loc[mouse_choice_idx, "L_prob"] = choice_prob[0, :]
        df_trials_fm.loc[mouse_choice_idx, "R_prob"] = choice_prob[1, :]
        df_trials_fm.loc[mouse_choice_idx, "Q_left"] = qvals[0, :-1]
        df_trials_fm.loc[mouse_choice_idx, "Q_right"] = qvals[1, :-1]

        if "CK" in model_name:
            df_trials_fm.loc[mouse_choice_idx, "L_kernel"] = choice_kernel[0, :-1]
            df_trials_fm.loc[mouse_choice_idx, "R_kernel"] = choice_kernel[1, :-1]

    df_sess_params = pd.DataFrame(df_sess_params)
    df_sess_fm = df_sess.merge(df_sess_params, how="left", on=["ses_idx"])

    if loc:
        df_sess_fm.to_csv(loc + "df_sess_fm.csv")
        df_trials_fm.to_csv(loc + "df_trials_fm.csv")

    else:
        return df_trials_fm, df_sess_fm



"""Testing for utility functions for processing dynamic foraging data."""

import unittest

import numpy as np
import pandas as pd
import glob

from src.aind_dynamic_foraging_data_utils import alignment, nwb_utils


class DynamicForagingTest(unittest.TestCase):
    """
    A class for testing the dynamic_foraging_utils module.
    """

    def test_create_df_session(self):
        """
        tests the `create_df_session` function
        """
        nwb_files = glob.glob("./tests/nwb/**.nwb")
        # create a dataframe for each nwb file

        df = nwb_utils.create_df_session(nwb_files)
        assert len(df) == len(nwb_files)

        df = nwb_utils.create_df_session(nwb_files[0])
        assert len(df) == 1

    def test_create_df_trials(self):
        """
        tests the `create_df_trials` function
        """
        nwb_files = glob.glob("./tests/nwb/**.nwb")
        # create a dataframe for one nwb file
        df = nwb_utils.create_df_trials(nwb_files[0])
        # check that the dataframe has correct session names
        session_name = "_".join(nwb_files[0].split("/")[-1].split("_")[0:-1])
        assert df.ses_idx[0] == session_name, "{} != {}".format(session_name, df.ses_idx[0])

    def test_get_time_array_with_sampling_rate(self):
        """
        tests the `get_time_array` function while passing
        a sampling_rate argument
        """
        # this should give [-1, 1) with steps of 0.5, exclusive of the endpoint
        t_array = alignment.get_time_array(
            t_start=-1, t_end=1, sampling_rate=2, include_endpoint=False
        )
        assert (t_array == np.array([-1.0, -0.5, 0.0, 0.5])).all()

        # this should give [-1, 1] with steps of 0.5, inclusive of the endpoint
        t_array = alignment.get_time_array(
            t_start=-1, t_end=1, sampling_rate=2, include_endpoint=True
        )
        assert (t_array == np.array([-1.0, -0.5, 0.0, 0.5, 1.0])).all()

        # this should give [-1, 0.75) with steps of 0.5.
        # becuase the desired range (1.75) is not evenly divisible by the
        # step size (0.5), the array should end before the desired endpoint
        t_array = alignment.get_time_array(
            t_start=-1, t_end=0.75, sampling_rate=2, include_endpoint=False
        )
        assert (t_array == np.array([-1.0, -0.5, 0.0, 0.5])).all()

        # this should give [-1, 0.75) with steps of 0.5.
        # becuase the desired range (1.75) is not evenly divisible by the
        # step size (0.5), the array should end before the desired endpoint
        t_array = alignment.get_time_array(
            t_start=-1, t_end=0.75, sampling_rate=2, include_endpoint=True
        )
        assert (t_array == np.array([-1.0, -0.5, 0.0, 0.5])).all()

    def test_get_time_array_with_step_size(self):
        """
        tests the `get_time_array` function while passing a step_size argument
        """
        # this should give [-1, 1) with steps of 0.5, exclusive of the endpoint
        t_array = alignment.get_time_array(
            t_start=-1, t_end=1, step_size=0.5, include_endpoint=False
        )
        assert (t_array == np.array([-1.0, -0.5, 0.0, 0.5])).all()

        # this should give [-1, 1] with steps of 0.5, inclusive of the endpoint
        t_array = alignment.get_time_array(
            t_start=-1, t_end=1, step_size=0.5, include_endpoint=True
        )
        assert (t_array == np.array([-1.0, -0.5, 0.0, 0.5, 1.0])).all()

        # this should give [-1, 0.75) with steps of 0.5.
        # becuase the desired range (1.75) is not evenly
        # divisible by the step size (0.5), the array should
        # end before the desired endpoint
        t_array = alignment.get_time_array(
            t_start=-1, t_end=0.75, step_size=0.5, include_endpoint=False
        )
        assert (t_array == np.array([-1.0, -0.5, 0.0, 0.5])).all()

        # this should give [-1, 0.75) with steps of 0.5.
        # becuase the desired range (1.75) is not evenly
        # divisible by the step size (0.5),
        # the array should end before the desired endpoint
        t_array = alignment.get_time_array(
            t_start=-1, t_end=0.75, step_size=0.5, include_endpoint=True
        )
        assert (t_array == np.array([-1.0, -0.5, 0.0, 0.5])).all()

    def test_event_triggered_response(self):
        """
        tests the `test_event_triggered_response` function
        """
        # make a time vector from -10 to 110
        t = np.arange(-10, 110, 0.01)

        # Make a dataframe with one column as time, and another
        # column called 'sinusoid' defined as sin(2*pi*t)
        # The sinusoid column will have a period of 1
        df = pd.DataFrame({"time": t, "sinusoid": np.sin(2 * np.pi * t)})
        df_copy = df.copy(deep=True)

        # Make an event triggered response
        etr = alignment.event_triggered_response(
            data=df,
            t="time",
            y="sinusoid",
            event_times=np.arange(100),
            t_before=1,
            t_after=1,
            output_sampling_rate=100,
        )

        # Assert that the average value of the agrees with expectations
        assert np.isclose(etr.query("time == 0")["sinusoid"].mean(), 0, rtol=0.01)
        assert np.isclose(etr.query("time == 0.25")["sinusoid"].mean(), 1, rtol=0.01)
        assert np.isclose(etr.query("time == 0.5")["sinusoid"].mean(), 0, rtol=0.01)
        assert np.isclose(etr.query("time == 0.75")["sinusoid"].mean(), -1, rtol=0.01)
        assert np.isclose(etr.query("time == 1")["sinusoid"].mean(), 0, rtol=0.01)

        # Assert that the dataframe is unchanged
        pd.testing.assert_frame_equal(df, df_copy)

    def test_event_triggered_response_censor(self):
        """
        tests the censoring property of the event_triggered_response function
        """
        # make a sample test set
        t = np.arange(0, 10, 0.01)
        y = np.array([10] * len(t))
        event_times = [2, 3, 4, 5, 6, 7, 8]
        for e in event_times:
            y[(t > e) & (t < (e + 0.25))] = 1

        df = pd.DataFrame({"time": t, "y": y})

        # make etr
        etr_censored = alignment.event_triggered_response(
            data=df,
            t="time",
            y="y",
            event_times=event_times,
            t_before=2,
            t_after=2,
            output_sampling_rate=100,
            censor=True,
        )

        # assert properties of etr
        assert np.isclose(etr_censored.query("time > 1")["y"].mean(), 10, rtol=0.01)
        assert np.isclose(etr_censored.query("time < -1")["y"].mean(), 10, rtol=0.01)

        # Test with arbitrary list of censor times
        censor_times = [2, 2.5, 3, 3.5, 4, 4.5, 5, 5.5, 6, 6.5, 7, 7.5, 8]
        etr_censored = alignment.event_triggered_response(
            data=df,
            t="time",
            y="y",
            event_times=event_times,
            t_before=2,
            t_after=2,
            output_sampling_rate=100,
            censor=True,
            censor_times=censor_times,
        )

        # assert properties of etr
        assert np.isclose(etr_censored.query("time > 1")["y"].mean(), 10, rtol=0.01)
        assert np.isclose(etr_censored.query("time < 0")["y"].mean(), 10, rtol=0.01)

    def test_event_triggered_response_nan_policy(self):
        """
        tests the `test_event_triggered_response` function
        """
        # make a time vector from -10 to 110
        t = np.arange(-10, 110, 0.01)

        # Make a dataframe with one column as time, and another
        # column called 'sinusoid' defined as sin(2*pi*t)
        # The sinusoid column will have a period of 1
        df = pd.DataFrame({"time": t, "sinusoid": np.sin(2 * np.pi * t)})

        # Make an event triggered response, NaN values are outside window
        df.loc[0:100, "sinusoid"] = np.nan
        etr = alignment.event_triggered_response(
            data=df,
            t="time",
            y="sinusoid",
            event_times=np.arange(100),
            t_before=1,
            t_after=1,
            output_sampling_rate=100,
            nan_policy="error",
        )

        # Raises an error
        df.loc[3980:4050, "sinusoid"] = np.nan
        with self.assertRaises(Exception):
            alignment.event_triggered_response(
                data=df,
                t="time",
                y="sinusoid",
                event_times=np.arange(100),
                t_before=1,
                t_after=1,
                output_sampling_rate=100,
                nan_policy="error",
            )

        # outputs are NaNs around NaN data
        etr = alignment.event_triggered_response(
            data=df,
            t="time",
            y="sinusoid",
            event_times=np.arange(100),
            t_before=1,
            t_after=1,
            output_sampling_rate=100,
            nan_policy="exclude",
        )
        assert np.isclose(
            etr.query("event_number == 29")["sinusoid"].isnull().mean(), 1.0, rtol=0.01
        )
        assert np.isclose(
            etr.query("event_number == 30")["sinusoid"].isnull().mean(), 1.0, rtol=0.01
        )
        assert np.isclose(
            etr.query("event_number == 31")["sinusoid"].isnull().mean(), 1.0, rtol=0.01
        )
        assert np.isclose(
            etr.query("event_number not in [29,30,31]")["sinusoid"].isnull().mean(), 0.0, rtol=0.01
        )

        # outputs are NaNs around NaN data
        etr = alignment.event_triggered_response(
            data=df,
            t="time",
            y="sinusoid",
            event_times=np.arange(100),
            t_before=1,
            t_after=1,
            output_sampling_rate=100,
            nan_policy="interpolate",
        )
        assert np.isclose(etr["sinusoid"].isnull().mean(), 0.0, rtol=0.01)


if __name__ == "__main__":
    unittest.main()



"""
Utility functions for processing dynamic foraging data.
    load_nwb_from_filename
    create_single_df_session_inner
    create_df_session
    create_single_df_session
    create_df_trials
    create_df_events
    create_df_fip
"""

import os
import re
import warnings

import numpy as np
import pandas as pd
from hdmf_zarr import NWBZarrIO
from pynwb import NWBHDF5IO
from datetime import date

# If we adjust time_in_session, adjust it to this
SESSION_ALIGNMENT = "goCue_start_time"

# Tolerance for responses to be outside the response window
RESPONSE_TIMING_TOLERANCE = 0.005

# Tolerance for responses before the go cue
CHOICE_TIMING_TOLERANCE = 0.005


def load_nwb_from_filename(filename):
    """
    Load NWB from file, checking for HDF5 or Zarr
    if filename is not a string, then return the input, assuming its the NWB file already
    """

    if type(filename) is str:
        if os.path.isdir(filename) or (filename.startswith("s3://") and filename.endswith(".nwb")):
            io = NWBZarrIO(filename, mode="r")
            nwb = io.read()
            return nwb
        elif os.path.isfile(filename):
            io = NWBHDF5IO(filename, mode="r")
            nwb = io.read()
            return nwb
        else:
            raise FileNotFoundError(filename)
    else:
        # Assuming its already an NWB
        return filename


def create_single_df_session_inner(nwb):
    """
    given a nwb file, output a tidy dataframe
    """
    df_trials = nwb.trials.to_dataframe()

    # Reformat data
    choice_history = df_trials.animal_response.map({0: 0, 1: 1, 2: np.nan}).values
    reward_history = np.vstack([df_trials.rewarded_historyL, df_trials.rewarded_historyR])

    # -- Session-based table --
    # - Meta data -
    session_start_time_from_meta = nwb.session_start_time
    session_date_from_meta = session_start_time_from_meta.strftime("%Y-%m-%d")
    subject_id_from_meta = nwb.subject.subject_id

    # TODO, should reprocess old files, and remove this logic
    if "behavior" in nwb.session_id:
        splits = nwb.session_id.split("_")
        subject_id = splits[1]
        session_date = splits[2]
        nwb_suffix = splits[3].replace("-", "")
    else:
        old_re = re.match(
            r"(?P<subject_id>\d+)_(?P<date>\d{4}-\d{2}-\d{2})(?:_(?P<n>\d+))?\.json",
            nwb.session_id,
        )

        if old_re is not None:
            # If there are more than one "bonsai sessions" (the trainer clicked "Save" button in the
            # GUI more than once) in a certain day,
            # parse nwb_suffix from the file name (0, 1, 2, ...)
            subject_id, session_date, nwb_suffix = old_re.groups()
            nwb_suffix = int(nwb_suffix) if nwb_suffix is not None else 0
        else:
            # After https://github.com/AllenNeuralDynamics/dynamic-foraging-task/commit/
            # 62d0e9e2bb9b47a8efe8ecb91da9653381a5f551,
            # the suffix becomes the session start time. Therefore, I use HHMMSS as the nwb suffix,
            # which still keeps the order as before.

            # Typical situation for multiple bonsai sessions per day is that the
            # RAs pressed more than once "Save" button but only started the session once.
            # Therefore, I should generate nwb_suffix from the bonsai file name
            # instead of session_start_time.
            subject_id, session_date, session_json_time = re.match(
                r"(?P<subject_id>\d+)_(?P<date>\d{4}-\d{2}-\d{2})(?:_(?P<time>.*))\.json",
                nwb.session_id,
            ).groups()
            nwb_suffix = int(session_json_time.replace("-", ""))

    # Ad-hoc bug fixes for some mistyped mouse ID
    if subject_id in ("689727"):
        subject_id_from_meta = subject_id

    assert subject_id == subject_id_from_meta, (
        f"Subject name from the metadata ({subject_id_from_meta}) does not match "
        f"that from json name ({subject_id})!!"
    )
    assert session_date == session_date_from_meta, (
        f"Session date from the metadata ({session_date_from_meta}) does not match "
        f"that from json name ({session_date})!!"
    )

    session_index = pd.MultiIndex.from_tuples(
        [(subject_id, session_date, nwb_suffix)],
        names=["subject_id", "session_date", "nwb_suffix"],
    )

    # Parse meta info
    # TODO: when generating nwb, put meta info in nwb.scratch and get rid of the regular expression
    # This could be significantly cleaned up based on new metadata format
    # But im making it consistent for now
    # TODO, should reprocess old files, and remove this logic
    if "behavior" not in nwb.session_id:
        extra_water, rig = re.search(
            r"Give extra water.*:(\d*(?:\.\d+)?)? .*?(?:tower|box):(.*)?",
            nwb.session_description,
        ).groups()
        weight_after_session = re.search(
            r"Weight after.*:(\d*(?:\.\d+)?)?", nwb.subject.description
        ).groups()[0]

        extra_water = float(extra_water) if extra_water != "" else 0
        weight_after_session = float(weight_after_session) if weight_after_session != "" else np.nan
        weight_before_session = float(nwb.subject.weight) if nwb.subject.weight != "" else np.nan
        user_name = nwb.experimenter[0]
    elif nwb.scratch:
        rig = nwb.scratch["metadata"].box[0]
        user_name = nwb.experimenter
        weight_after_session = nwb.scratch["metadata"].weight_after[0]
        water_during_session = nwb.scratch["metadata"].water_in_session_total[0]
        weight_before_session = weight_after_session - water_during_session
        extra_water = nwb.scratch["metadata"].water_in_session_manual[0]
    else:
        rig = 0
        user_name = 0
        weight_after_session = 0
        water_during_session = 0
        weight_before_session = 0
        extra_water = 0

    dict_meta = {
        "rig": rig,
        "user_name": user_name,
        "experiment_description": nwb.experiment_description,
        "task": nwb.protocol,
        "session_start_time": session_start_time_from_meta,
        "weight_before_session": weight_before_session,
        "weight_after_session": weight_after_session,
        "water_during_session": weight_after_session - weight_before_session,
        "water_extra": extra_water,
    }

    df_session = pd.DataFrame(
        dict_meta,
        index=session_index,
    )
    # Use hierarchical index (type = {'metadata', 'session_stats'}, variable = {...}, etc.)
    df_session.columns = pd.MultiIndex.from_product(
        [["metadata"], dict_meta.keys()], names=["type", "variable"]
    )

    # - Compute session-level stats -
    # TODO: Ideally, all these simple stats could be computed in the GUI, and
    # the GUI sends a copy to the meta session.json file and to the nwb file as well.

    total_trials = len(df_trials)
    finished_trials = np.sum(~np.isnan(choice_history))
    reward_trials = np.sum(reward_history)

    reward_rate = reward_trials / finished_trials

    # TODO: add more stats
    # See code here: https://github.com/AllenNeuralDynamics/map-ephys/blob/
    # 7a06a5178cc621638d849457abb003151f7234ea/pipeline/foraging_analysis.py#L70C8-L70C8
    # early_lick_ratio =
    # double_dipping_ratio =
    # block_num
    # mean_block_length
    # mean_reward_sum
    # mean_reward_contrast
    # autowater_num
    # autowater_ratio
    #
    # mean_iti
    # mean_reward_sum
    # mean_reward_contrast
    # ...

    # foraging_eff_func = (
    #    foraging_eff_baiting if "bait" in nwb.protocol.lower() else foraging_eff_no_baiting
    # )
    # foraging_eff, foraging_eff_random_seed = foraging_eff_func(
    #    reward_rate, p_reward[LEFT, :], p_reward[RIGHT, :]
    # )

    # -- Add session stats here --
    dict_session_stat = {
        "total_trials": total_trials,
        "finished_trials": finished_trials,
        "finished_rate": finished_trials / total_trials,
        "ignore_rate": np.sum(np.isnan(choice_history)) / total_trials,
        "reward_trials": reward_trials,
        "reward_rate": reward_rate,
        # TODO: add more stats here
    }

    # Generate df_session_stat
    df_session_stat = pd.DataFrame(dict_session_stat, index=session_index)
    df_session_stat.columns = pd.MultiIndex.from_product(
        [["session_stats"], dict_session_stat.keys()],
        names=["type", "variable"],
    )

    # -- Add automatic training --
    if "auto_train_engaged" in df_trials.columns:
        df_session["auto_train", "curriculum_name"] = (
            np.nan
            if df_trials.auto_train_curriculum_name.mode()[0] == "none"
            else df_trials.auto_train_curriculum_name.mode()[0]
        )
        df_session["auto_train", "curriculum_version"] = (
            np.nan
            if df_trials.auto_train_curriculum_version.mode()[0] == "none"
            else df_trials.auto_train_curriculum_version.mode()[0]
        )
        df_session["auto_train", "curriculum_schema_version"] = (
            np.nan
            if df_trials.auto_train_curriculum_schema_version.mode()[0] == "none"
            else df_trials.auto_train_curriculum_schema_version.mode()[0]
        )
        df_session["auto_train", "current_stage_actual"] = (
            np.nan
            if df_trials.auto_train_stage.mode()[0] == "none"
            else df_trials.auto_train_stage.mode()[0]
        )
        df_session["auto_train", "if_overriden_by_trainer"] = (
            np.nan
            if all(df_trials.auto_train_stage_overridden.isna())
            else df_trials.auto_train_stage_overridden.mode()[0]
        )

        # Add a flag to indicate whether any of the auto train settings were changed
        # during the training
        df_session["auto_train", "if_consistent_within_session"] = (
            len(df_trials.groupby([col for col in df_trials.columns if "auto_train" in col])) == 1
        )
    else:
        for field in [
            "curriculum_name",
            "curriculum_version",
            "curriculum_schema_version",
            "current_stage_actual",
            "if_overriden_by_trainer",
        ]:
            df_session["auto_train", field] = None

    # -- Merge to df_session --
    df_session = pd.concat([df_session, df_session_stat], axis=1)
    return df_session


def create_df_session(nwb_filename):
    """
    Creates a dataframe where each row is a session
    nwb_filename can be either a single nwb file, a single filepath
    or a list of nwb files, or a list of nwb filepaths
    """
    if (type(nwb_filename) is not str) and (hasattr(nwb_filename, "__iter__")):
        dfs = []
        for nwb_file in nwb_filename:
            dfs.append(create_single_df_session(nwb_file))
        return pd.concat(dfs)
    else:
        return create_single_df_session(nwb_filename)


# % Process nwb and create df_session for every single session
def create_single_df_session(nwb_filename):
    """
    create a dataframe for a single session
    """
    nwb = load_nwb_from_filename(nwb_filename)

    df_session = create_single_df_session_inner(nwb)

    df_session.columns = df_session.columns.droplevel("type")
    df_session = df_session.reset_index()
    df_session["ses_idx"] = (
        df_session["subject_id"].values + "_" + df_session["session_date"].values
    )
    df_session = df_session.rename(columns={"variable": "session_num"})
    return df_session


def create_df_trials(nwb_filename, adjust_time=True, verbose=True):  # NOQA C901
    """
    Process nwb and create df_trials for every single session

    ARGS:
    nwb_filename (str or NWB object), the session to extract the trials from
    adjust_time (bool) if true, adjust t0 to be the first gocue

    RETURNS:
    A pandas dataframe containing the columns of nwb.trials plus:
    "_in_trial" time alignments where time is relative to the go cue on that trial
    "_in_session" time alignments where time is relative to the first go cue
        of the session.
    earned_reward, (bool) whether a reward was earned in that trial
    extra_reward (bool) whether a manual reward was given in that trial
    """

    # If we are given a filename, load the NWB object itself
    nwb = load_nwb_from_filename(nwb_filename)

    # Parse subject and session_date
    if nwb.session_id.startswith("behavior") or nwb.session_id.startswith("FIP"):
        splits = nwb.session_id.split("_")
        subject_id = splits[1]
        session_date = splits[2]
    else:
        splits = nwb.session_id.split("_")
        subject_id = splits[0]
        session_date = splits[1]
    ses_idx = subject_id + "_" + session_date

    # Build dataframe
    df = nwb.trials.to_dataframe().reset_index()
    df = df.rename(columns={"id": "trial"})
    df["ses_idx"] = ses_idx

    # Adjust for gaps in trial start/stop, and use the last stop time
    last_stop = df.iloc[-1]["stop_time"]
    df["stop_time"] = df["start_time"].shift(-1, fill_value=last_stop)
    df = df.rename(columns={"start_time": "bonsai_start_time", "stop_time": "bonsai_stop_time"})

    # We skip these columns because they are how long the valve is open
    # not the times at which the valves were opened
    skip_cols = ["right_valve_open_time", "left_valve_open_time"]

    # compute times relative to start of trial and start of session
    t0 = nwb.trials[SESSION_ALIGNMENT][0]
    drop_cols = []
    for col in df.columns:
        if ("time" in col) and (col not in skip_cols):
            # Adjust all times relative to start of the first go cue
            if adjust_time:
                df[col + "_in_session"] = df[col] - t0
            else:
                df[col + "_in_session"] = df[col]

            # Adjust times relative to go cue on each trial
            if ("time" in col) and (col not in skip_cols):
                # Here we always align to goCue_start_time, not SESSION_ALIGNMENT
                # since this aligns events relative to the trial go cue, not the start
                # of the session
                df[col + "_in_trial"] = df[col].values - df["goCue_start_time"].values

            # Clean up these column names that are not clear
            drop_cols.append(col)

    # Add a column of raw time so users can map if they want
    df[SESSION_ALIGNMENT + "_raw"] = df[SESSION_ALIGNMENT]

    # Get lick and reward times
    key_from_acq = [
        "left_lick_time",
        "right_lick_time",
        "left_reward_delivery_time",
        "right_reward_delivery_time",
    ]
    if adjust_time:
        events = {key: nwb.acquisition[key].timestamps[:] - t0 for key in key_from_acq}
    else:
        events = {key: nwb.acquisition[key].timestamps[:] for key in key_from_acq}

    # Map events to trials
    # Here we map an event to the most recent goCue
    df["next_goCue_start_time_in_session"] = df["goCue_start_time_in_session"].shift(
        -1, fill_value=np.inf
    )
    drop_cols.append("next_goCue_start_time_in_session")

    # Create column with CHOICE TIMING TOLERANCE
    df["next_goCue_start_time_in_session_tolerance"] = (
        df["next_goCue_start_time_in_session"] - CHOICE_TIMING_TOLERANCE
    )
    df["goCue_start_time_in_session_tolerance"] = (
        df["goCue_start_time_in_session"] - CHOICE_TIMING_TOLERANCE
    )
    drop_cols.append("next_goCue_start_time_in_session_tolerance")
    drop_cols.append("goCue_start_time_in_session_tolerance")

    for event in key_from_acq:
        event_times = events[event]
        if event in ["left_lick_time", "right_lick_time"]:
            times = "_tolerance"
        else:
            times = ""
        df[event] = df.apply(
            lambda x: event_times[
                (event_times >= x["goCue_start_time_in_session" + times])
                & (event_times < x["next_goCue_start_time_in_session" + times])
            ],
            axis=1,
        )

    # Compute time of reward for each trial
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", message="All-NaN slice encountered")
        df["reward_time_in_session"] = df.apply(
            lambda x: np.nanmin(
                np.concatenate(
                    [
                        [np.nan],
                        x["right_reward_delivery_time"],
                        x["left_reward_delivery_time"],
                    ]
                )
            ),
            axis=1,
        )
    df["reward_time_in_trial"] = (
        df["reward_time_in_session"] - df[SESSION_ALIGNMENT + "_in_session"]
    )

    # Add annotation of reward types
    for event in ["right_reward_delivery_time", "left_reward_delivery_time"]:
        times = events[event]
        data = nwb.acquisition[event].data[:]
        mapper = {x[0]: x[1] for x in zip(times, data)}
        df[event.split("delivery_time")[0] + "type"] = [
            mapper[x[0]] if len(x) > 0 else np.nan for x in df[event]
        ]

    # Compute boolean of whether animal was rewarded
    # AutoWater and manual water is not included in earned_reward
    df["earned_reward"] = df.rewarded_historyR | df.rewarded_historyL
    if np.sum(df["right_reward_type"] == "earned") > 0:
        # We have reward type annotations, checking for backwards compatability
        df["extra_reward"] = (
            (df["right_reward_type"] == "manual")
            | (df["left_reward_type"] == "manual")
            | (df["right_reward_type"] == "auto")
            | (df["left_reward_type"] == "auto")
        )
    else:
        df["extra_reward"] = (~df["earned_reward"]) & df["reward_time_in_session"].notnull()

    # Compute time of choice for each trials
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", message="All-NaN slice encountered")
        df["choice_time_in_session"] = df.apply(
            lambda x: np.nanmin(
                np.concatenate([[np.nan], x["right_lick_time"], x["left_lick_time"]])
            ),
            axis=1,
        )
    df["choice_time_in_trial"] = (
        df["choice_time_in_session"] - df[SESSION_ALIGNMENT + "_in_session"]
    )

    # Filtering out choices greater than response window
    slow_choice = (
        df["choice_time_in_trial"] > df["response_duration"] + RESPONSE_TIMING_TOLERANCE
    ) & (~df["earned_reward"])
    df.loc[slow_choice, "choice_time_in_session"] = np.nan
    df.loc[slow_choice, "choice_time_in_trial"] = np.nan
    if np.sum(df["choice_time_in_trial"] > df["response_duration"] + RESPONSE_TIMING_TOLERANCE) > 0:
        warnings.warn("Response time greater than minimum, something unusual happened")

    # Sanity checks

    # prior to 2025/1/1, we did not include manual reward information.
    # This is a conservative estimate.
    manual_reward_date_cutoff = date(2025, 1, 1)

    rewarded_df = df.query("earned_reward")
    if not np.isnan(rewarded_df["reward_time_in_session"]).sum() == 0:
        if date.fromisoformat(session_date) <= manual_reward_date_cutoff:
            warnings.warn(
                "Rewarded trials without reward time. \
                This is likely due to manual rewards not being recorded in sessions from 2024"
            )
        else:
            raise AssertionError("Rewarded trials without reward time")

    assert (
        np.isnan(rewarded_df["choice_time_in_session"]).sum() == 0
    ), "Rewarded trials without choice time"

    earned_df = rewarded_df.query("not extra_reward")
    if not np.all(earned_df["choice_time_in_session"] <= earned_df["reward_time_in_session"]):
        if date.fromisoformat(session_date) <= manual_reward_date_cutoff:
            warnings.warn(
                "Reward before choice time. \
                This is likely due to manual rewards not being recorded in sessions from 2024"
            )
        else:
            raise AssertionError("Reward before choice time")

    assert np.all(
        rewarded_df["choice_time_in_trial"] >= -CHOICE_TIMING_TOLERANCE
    ), "Rewarded trial with negative choice_time_in_trial"

    check_rew_time = np.isnan(
        df.query("not earned_reward").query("not extra_reward")["reward_time_in_session"]
    )
    if not np.all(check_rew_time):
        if date.fromisoformat(session_date) <= manual_reward_date_cutoff:
            warnings.warn(
                "Unrewarded trials with reward time. If this was data from 2024, \
                        this is likely because extra_rewards are not recorded",
                UserWarning,
            )
        else:
            raise AssertionError("Unrewarded trials with reward time")

    # Drop columns
    drop_cols += key_from_acq
    df = df.drop(columns=drop_cols)

    if adjust_time and verbose:
        print(
            "Timestamps are adjusted such that `_in_session` timestamps start at the first go cue"
        )

    # Previously lickspout y coordinates were tied, so older data only reported one coordinate
    # with the adoption of the AIND lickspout stage, we migrated to y1 and y2 coordinates.
    # older NWB files should be reprocessed to ensure both coordinates are present
    if ("lickspout_position_y" in df) and ("lickspout_position_y1" not in df):
        text = (
            "Independent lickspout y coordinates are not provided. "
            "This DOES NOT indicate a data error, since older data "
            "did not allow independent y coordinate movement. It DOES "
            "indicate this NWB file needs to be reprocessed. Please report to "
            "https://github.com/AllenNeuralDynamics/aind-dynamic-foraging-data-utils/issues/67"
        )

        warnings.warn(text, UserWarning)

    return df


def create_df_events(nwb_filename, adjust_time=True, verbose=True):
    """
    returns a tidy dataframe of the events in the nwb file

    adjust_time (bool), set time of first goCue to t=0
    verbose (bool), give warnings for adjustments
    """

    nwb = load_nwb_from_filename(nwb_filename)

    # Build list of all event types in acqusition, ignore FIP events (no need for processing folder)
    event_types = set(nwb.acquisition.keys())

    channels = ["G", "R", "Iso"]
    fibers = ["0", "1", "2", "3", "4"]
    FIP_prefixes = [f"{c}_{f}" for c in channels for f in fibers]

    # Filter out all fibers
    event_types = {
        k for k in event_types if not any(k.startswith(prefix) for prefix in FIP_prefixes)
    }

    event_types -= {"FIP_falling_time", "FIP_rising_time"}

    # Determine time 0 as first go Cue
    if adjust_time:
        t0 = nwb.trials[SESSION_ALIGNMENT][0]
    else:
        t0 = 0

    # Iterate over event types and build a dataframe of each
    events = []
    for e in event_types:
        # For each event, get timestamps, data, and label
        raw_stamps = nwb.acquisition[e].timestamps[:]
        data = nwb.acquisition[e].data[:]
        labels = [e] * len(data)
        stamps = raw_stamps - t0
        df = pd.DataFrame(
            {"timestamps": stamps, "data": data, "event": labels, "raw_timestamps": raw_stamps}
        )
        events.append(df)

    # Add keys from trials table
    trial_events = ["goCue_start_time"]
    for e in trial_events:
        raw_stamps = nwb.trials[:][e].values
        labels = [e] * len(raw_stamps)
        data = [1] * len(raw_stamps)
        stamps = raw_stamps - t0
        df = pd.DataFrame(
            {"timestamps": stamps, "data": data, "event": labels, "raw_timestamps": raw_stamps}
        )
        events.append(df)

    # Build dataframe by concatenating each event
    df = pd.concat(events)
    df = df.sort_values(by="timestamps")
    df = df.dropna(subset="timestamps").reset_index(drop=True)

    # Add trial index for each event
    trial_starts = nwb.trials.goCue_start_time[:] - t0
    last_stop = np.inf
    trial_index = []
    for index, e in df.iterrows():
        starts = np.where(e.timestamps >= trial_starts)[0]
        if len(starts) == 0:
            trial_index.append(-1)
        elif e.timestamps > last_stop:
            trial_index.append(len(trial_starts))
        else:
            trial_index.append(starts[-1])
    df["trial"] = trial_index

    # Sanity check that the first go cue is time 0
    gocues = df.query("event == @SESSION_ALIGNMENT")
    if (len(gocues) > 0) and (adjust_time):
        assert np.isclose(gocues.iloc[0]["timestamps"], 0, rtol=0.01)
    # TODO, need more checks here for time alignment on trial index.

    if adjust_time and verbose:
        print(
            "Timestamps are adjusted such that `_in_session` timestamps start at the first go cue"
        )
    return df


def create_df_fip(nwb_filename, tidy=True, adjust_time=True, verbose=True):
    """
    returns a dataframe of the FIB data in the nwb file
    if tidy, return a tidy dataframe
    if not tidy, return pivoted by timestamp

    adjust_time (bool), set time of first goCue to t=0
    """

    nwb = load_nwb_from_filename(nwb_filename)

    # Build list of all FIB events in NWB file
    nwb_data = nwb.acquisition
    if len(nwb.processing):
        nwb_data = nwb.acquisition | nwb.processing["fiber_photometry"].data_interfaces

    event_types = set(nwb_data.keys())

    channels = ["G", "R", "Iso"]
    fibers = ["0", "1", "2", "3", "4"]
    FIP_prefixes = [f"{c}_{f}" for c in channels for f in fibers]

    # Filter out all fibers
    event_types = {k for k in event_types if any(k.startswith(prefix) for prefix in FIP_prefixes)}

    event_types.add("FIP_falling_time")
    event_types.add("FIP_rising_time")

    # If no FIB data available
    if len(event_types) == 0:
        return None

    # Determine time 0 as first go Cue
    if adjust_time:
        t0 = nwb.trials[SESSION_ALIGNMENT][0]
    else:
        t0 = 0

    # Iterate over event types and build a dataframe of each
    events = []
    for e in event_types:
        # For each event, get timestamps, data, and label
        raw_stamps = nwb_data[e].timestamps[:]
        data = nwb_data[e].data[:]
        labels = [e] * len(data)
        stamps = raw_stamps - t0
        df = pd.DataFrame(
            {"timestamps": stamps, "data": data, "event": labels, "raw_timestamps": raw_stamps}
        )
        events.append(df)

    # Build dataframe by concatenating each event
    df = pd.concat(events)
    df = df.sort_values(by="timestamps")
    df = df.dropna(subset="timestamps").reset_index(drop=True)

    # Add session_idx with subject ID and session date info - JL
    if nwb.session_id.startswith("behavior") or nwb.session_id.startswith("FIP"):
        splits = nwb.session_id.split("_")
        subject_id = splits[1]
        session_date = splits[2]
    else:
        splits = nwb.session_id.split("_")
        subject_id = splits[0]
        session_date = splits[1]
    ses_idx = subject_id + "_" + session_date
    df["ses_idx"] = ses_idx

    if adjust_time and verbose:
        print(
            "Timestamps are adjusted such that `_in_session` timestamps start at the first go cue"
        )

    # pivot table based on timestamps
    if not tidy:
        df_pivoted = pd.pivot(df, index="timestamps", columns=["event"], values="data")
        return df_pivoted
    else:
        return df



"""
    This is an example script to extract dataframes
    from an NWB file
"""

import glob

from aind_dynamic_foraging_data_utils import nwb_utils

# List of filepaths to NWBs
NWB_FILES = glob.glob("../tests/nwb/**.nwb")

# Load a single NWB file
nwb = nwb_utils.load_nwb_from_filename(NWB_FILES[0])

# Make dataframe of sessions, each row is a session
# Accepts: NWB objects, NWB files, list of NWB files
df_session = nwb_utils.create_df_session(NWB_FILES)

# Dataframe of trials, each row is a trial
# Accepts NWB filepath or object
df_trials = nwb_utils.create_df_trials(nwb)
df_trials = nwb_utils.create_df_trials(NWB_FILES[0])

# tidy data frame of all events in the session
# Accepts NWB filepath or object
df_events = nwb_utils.create_df_events(nwb)
df_events = nwb_utils.create_df_events(NWB_FILES[0])

# Data frame of FIB data
# Accepts NWB filepath or object
df_fip = nwb_utils.create_df_fip(nwb)
df_fip = nwb_utils.create_df_fip(NWB_FILES[0])



"""
Important utility functions for enriching the dataframes

    extract_time_window
    extract_signal_window
    enrich_fip_in_df_trials
    zscore_fip
    tidy_df_trials_fip
    remove_tonic_df_fip
    enrich_df_trials_fm
"""

import itertools
import warnings

import numpy as np
import pandas as pd
from scipy.stats import zscore


def extract_time_window(
    trial_i, timepoints, alignment_events, offsets, absolute_time, align_in_trial
):
    """
    extract out time window around the offsets of alignment events
    Args:
        trial_i (pd.Series): row of trial data from df_trials
        timepoints (list): timestamps for a particular session
        alignment_events (list): 2 timepoints in trial to align the offsets to
        offsets (list): Offsets from the 2 timepoints to extract time window
        absolute_time (str): column name in df_trials that defines start of trial in session
        align_in_trial (Bool): return the time window in_trial (as opposed to in_session)
    Returns:
        window (list): list of timepoints given in_session or in_trial (if align_in_trial)
    """
    # t_start, t_end for the trial
    t_start = trial_i[alignment_events[0]] + offsets[0] + trial_i[absolute_time]
    t_end = trial_i[alignment_events[1]] + offsets[1] + trial_i[absolute_time]
    # window of timestamps in session time
    window = timepoints[(timepoints > t_start) & (timepoints < t_end)]
    return window - trial_i[absolute_time] if align_in_trial else window


def extract_signal_window(trial_i, timepoints, varpoints, alignment_events, offsets, absolute_time):
    """
    extract out signal for the time window around offsets around alignment events.
    Args:
        trial_i (pd.Series): row of trial data from df_trials
        timepoints (list): timestamps for a particular session
        varpoints (list): signal for a particular session
        alignment_events (list): 2 timepoints in trial to align the offsets to
        offsets (list): Offsets from the 2 timepoints to extract time window
        absolute_time (str): column name in df_trials that defines start of trial in session
    Returns:
        window (list): list of timepoints given in_session or in_trial (if align_in_trial)
    """
    # t_start, t_end for the trial
    t_start = trial_i[alignment_events[0]] + offsets[0] + trial_i[absolute_time]
    t_end = trial_i[alignment_events[1]] + offsets[1] + trial_i[absolute_time]
    return varpoints[(timepoints > t_start) & (timepoints < t_end)]


def enrich_fip_in_df_trials(df_fip, df_trials):
    """
    Processes df_fip and df_trials, computing z-scored data and align FIP data in df_trials
    Args:
        df_fip (pd.DataFrame): A DataFrame with FIP data, each row a timepoint for a signal
        df_trials (pd.DataFrame): A DataFrame of trials (each row a trial)
    Returns:
        df_fip_z (pd.DataFrame): df_fip updated with a data_z column for
                                z-scored data within a session
        df_fip_trials (pd.DataFrame):
                        df_trials updated with FIP data from df_fip
                        Each trial now has the timestamps and activity for
                        1 second before goCue to 1 second before
                        next goCue. new columns added as:
                        timestamps_in_session_{event}: timestamps in session time for signal 'event'
                        timestamps_in_trial_{event}: timestamps in trial time for signal 'event'
                        data_{event}: corresponding data for signal 'event'
                        data_z_{event}: corresponding z-scored data for signal 'event'

    """

    # get the goCue of the next trial defined
    last_timestamps = df_fip.groupby("ses_idx", sort=False)["timestamps"].last()
    df_trials["goCue_start_time_next_in_session"] = df_trials.groupby("ses_idx")[
        "goCue_start_time_in_session"
    ].shift(
        -1, fill_value=-1
    )  # noqa: E501
    df_trials.loc[
        df_trials["goCue_start_time_next_in_session"] == -1, "goCue_start_time_next_in_session"
    ] = last_timestamps  # noqa: E501
    df_trials["goCue_start_time_next_in_trial"] = (
        df_trials["goCue_start_time_next_in_session"] - df_trials["goCue_start_time_in_session"]
    )  # noqa: E501
    # set alignment events and offsets
    alignment_events = ["goCue_start_time_in_trial", "goCue_start_time_next_in_trial"]
    offsets = [-1, -1]

    # updated dataframes
    df_trials_fip = df_trials.copy()

    # z-score the data
    df_fip_z = zscore_fip(df_fip)

    for (ses_idx, event), df_fip_i in df_fip_z.groupby(["ses_idx", "event"]):
        # pull fip data into df_trials
        df_trials_ses = df_trials_fip.loc[df_trials_fip["ses_idx"] == ses_idx, :]
        if len(df_trials_ses) == 0:
            continue
        timepoints = df_fip_i["timestamps"].values
        absolute_time = "goCue_start_time_in_session"
        df_trials_fip.loc[df_trials_fip["ses_idx"] == ses_idx, f"timestamps_in_trial_{event}"] = (
            df_trials_ses.apply(
                lambda trial_i: extract_time_window(
                    trial_i,
                    timepoints,
                    alignment_events,
                    offsets,
                    absolute_time,
                    align_in_trial=True,
                ),
                axis=1,
            )
        )
        df_trials_fip.loc[df_trials_fip["ses_idx"] == ses_idx, f"timestamps_in_session_{event}"] = (
            df_trials_ses.apply(
                lambda trial_i: extract_time_window(
                    trial_i,
                    timepoints,
                    alignment_events,
                    offsets,
                    absolute_time,
                    align_in_trial=False,
                ),
                axis=1,
            )
        )
        for var_event in ["data", "data_z"]:
            varpoints = df_fip_i[var_event].values
            df_trials_fip.loc[df_trials_fip["ses_idx"] == ses_idx, f"{var_event}_{event}"] = (
                df_trials_ses.apply(
                    lambda trial_i: extract_signal_window(
                        trial_i, timepoints, varpoints, alignment_events, offsets, absolute_time
                    ),
                    axis=1,
                )
            )
    return (df_fip_z, df_trials_fip)


def zscore_fip(df_fip, data_col="data"):
    """
    z-score FIP signal separately for each channel and session
    df_fip: dataframe of FIP data, must contain signal in "data" column, channels
        separated by "event", and sessions separated by "ses_idx"
    data_col (str): name of the column in df_fip to be zscored
    returns
    df_fip_z: dataframe with additional column "<data_col>_z" with z-scored signal

    example:
    nwb.df_fip = zscore_fip(nwb.df_fip)
    """
    df_fip_z = df_fip.copy()
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=RuntimeWarning)
        # Ignore run-time warnings because FIP_falling_time and FIP_rising_time are all ones.
        df_fip_z.loc[:, "{}_z".format(data_col)] = df_fip_z.groupby(["ses_idx", "event"])[
            data_col
        ].transform(lambda x: zscore(x, ddof=1, nan_policy="omit"))
    return df_fip_z


def tidy_df_trials_fip(df_fip, df_trials_fip, col_prefix_signal="data"):
    """
    Converts df_trials_fip's columns of FIP data into a tidy long-format columns in df_fip
    - Extracts event names from `timestamps_{event}` columns.
    - Explodes from FIP columns `{col_prefix_signal}_*{event}_norm`
    - Merges the tidied data with df_fip on `ses_idx` and `timestamps`.

    Args:
        df_fip (pd.DataFrame): A DataFrame with FIP data, each row a timepoint for a signal
        df_trials (pd.DataFrame): A DataFrame of trials (each row a trial)
        col_prefix_signal (str, optional): The prefix of the signal column name. Defaults to 'data'.


    Returns:
        df_tidy (pd.DataFrame): df_fip with the untidy df_trials_fip baseline removed data
                                added in. Two columns for data_norm (baseline removed signal)
                                and data_z_norm (z-scored baseline removed signal).
    """
    # Step 1: Identify relevant timestamp and data columns
    timestamp_cols = [
        col for col in df_trials_fip.columns if col.startswith("timestamps_in_session")
    ]
    data_cols = [
        col
        for col in df_trials_fip.columns
        if col.startswith(col_prefix_signal + "_") and col.endswith("_norm")
    ]

    # Step 2: Initialize an empty list to store exploded DataFrames
    exploded_dfs = []

    # Step 3: Iterate over each timestamp-data column pair and explode them
    for timestamp_col in timestamp_cols:
        event = timestamp_col.replace("timestamps_in_session_", "")
        # Find matching data columns for this event
        matching_data_cols = [col for col in data_cols if event + "_norm" in col]

        # Create df_subset with timestamps and all matching data columns
        df_subset = df_trials_fip[["ses_idx", timestamp_col] + matching_data_cols].copy()
        df_subset = df_subset.rename(columns={timestamp_col: "timestamps"})
        df_subset["event"] = event  # Add event column

        # Rename data columns dynamically by removing the event part
        # Drop all rows with no signals
        rename_dict = {
            data_col: data_col.replace(f"_{event}_norm", "_norm") for data_col in matching_data_cols
        }
        df_subset = df_subset.rename(columns=rename_dict)
        df_subset = df_subset.dropna()

        # Explode timestamps and data together
        selected_cols = ["timestamps"] + [
            col for col in df_subset.columns if col.startswith(col_prefix_signal)
        ]
        df_subset = df_subset.explode(selected_cols, ignore_index=True)
        df_subset[selected_cols] = df_subset[selected_cols].astype(float)

        # Append to list for later concatenation
        exploded_dfs.append(df_subset)

    # Step 4: Concatenate all exploded DataFrames **before merging**
    df_exploded = pd.concat(exploded_dfs, ignore_index=True)

    # Step 5: Merge with df_fip on 'ses_idx' and 'timestamps'
    df_tidy = df_fip.merge(df_exploded, on=["ses_idx", "timestamps", "event"], how="left")
    return df_tidy.dropna().reset_index()


def remove_tonic_df_fip(
    df_fip,
    df_trials,
    df_trials_fip,
    col_prefix_signal="data",
    col_prefix_time="timestamps_in_trial",
):
    """
    Removes tonic activity by normalizing signal data against baseline.
    baseline is defined as the average activity 1 second before the goCue.

    Args:
        df_fip (pd.DataFrame): A DataFrame with FIP data, each row a timepoint for a signal
        df_trials (pd.DataFrame): A DataFrame of trials (each row a trial)
        df_trials_fip (pd.DataFrame): A DataFrame of trials with signal data per trial
        col_prefix_signal (str, optional): The prefix of the signal column name. Defaults to 'data'.
        col_prefix_time (str, optional): The prefix of the time column name.
                                            Defaults to 'timestamps_in_trial'.
    Returns:
        df_fip_tonic (pd.DataFrame): Updated df_fip with the baseline activity added as columns
                                'data_{event}_norm'.
        df_trials (pd.DataFrame): df_trials_fip with averaged baseline added as column
                                'data_{event}_baseline'.
        df_trials_fip (pd.DataFrame): df_trials_fip with baseline removed activity added
                                as column 'data_{sevent}_norm'.
    """
    # Step 1: Find all columns with data signals.
    # Skip any columns that already have baseline removed.
    col_signals = [
        col
        for col in df_trials_fip.columns
        if col.startswith(col_prefix_signal)
        and not col.endswith("_baseline")
        and not col.endswith("_norm")
    ]

    # Step 2: Loop through signal by signal, average the signal for baseline,
    # then remove the average activity per trial.

    # The baseline time range for normalization
    # Time range is given as offsets from goCue of current trial to
    # goCue of next trial. Defaults to [-1, 0].
    # Could be added as input argument in the future.
    offset_from_goCue = [-1, 0]
    for col_signal in col_signals:
        if col_signal + "_norm" in df_trials_fip.columns:
            continue
        signal_name = col_signal.removeprefix("data_z").removeprefix("data").removesuffix("_norm")
        # Calculate the baseline for each trial, add to df_trials and df_trials_fip
        signal_baseline = df_trials_fip.apply(
            lambda x: (
                np.nanmean(
                    x[col_signal][
                        (x[col_prefix_time + signal_name] < offset_from_goCue[1])
                        & (x[col_prefix_time + signal_name] > offset_from_goCue[0])
                    ]
                )
                if not np.isnan(x[col_signal]).all()
                else np.nan
            ),
            axis=1,
        )
        df_trials_fip.loc[:, col_signal + "_baseline"] = signal_baseline
        df_trials.loc[:, col_signal + "_baseline"] = signal_baseline
        # Normalize the signal by subtracting the baseline activity from the trial
        df_trials_fip.loc[:, col_signal + "_norm"] = df_trials_fip.apply(
            lambda x: (
                x[col_signal] - x[col_signal + "_baseline"]
                # Skip calculation if col_signal is NaN
                if not np.isnan(x[col_signal]).all()
                else np.nan
            ),
            axis=1,
        )
    df_fip_tonic = tidy_df_trials_fip(df_fip, df_trials_fip, col_prefix_signal)
    return (df_fip_tonic, df_trials, df_trials_fip)


def enrich_df_trials_fm(df_trials_fm):
    """
    enrich_df_trials_fm: enriches the df_trials_fm with additional columns like
                         RPE_earned, RPE_all, Q_chosen, Q_unchosen, Q_sum, Q_Delta, Q_change
                         model information can be found in repo: aind-dynamic-foraging-models
                         model fitting is done through this repo:
                         aind-analysis-arch-pipeine-dynamic-foraging
    RPE_earned: earned_reward - chosen_values
    RPE_all: (earned_reward + extra_reward) - chosen_values
    """
    df_trials_fm_enriched = pd.DataFrame()
    models = df_trials_fm["model_name"].unique()
    sessions = np.unique(df_trials_fm["ses_idx"])
    has_kernel = all([choice + "_kernel" in df_trials_fm.columns for choice in ["L", "R"]])
    for i_iter, (ses_idx, model_name) in enumerate(itertools.product(sessions[:], models)):
        df_ses = df_trials_fm[
            (df_trials_fm["ses_idx"] == ses_idx) & (df_trials_fm["model_name"] == model_name)
        ]  # noqa: E501
        choices = df_ses.choice.map({0: "L", 1: "R", 2: "I"}).values
        chosen_values = np.nan * np.zeros(len(df_ses))
        unchosen_values = np.nan * np.zeros(len(df_ses))
        chosen_probabilities = np.nan * np.zeros(len(df_ses))
        unchosen_probabilities = np.nan * np.zeros(len(df_ses))
        chosen_stay_probabilities = np.nan * np.zeros(len(df_ses))
        chosen_kernels = np.nan * np.zeros(len(df_ses))
        unchosen_kernels = np.nan * np.zeros(len(df_ses))
        # chosen_licks = np.nan * np.zeros(len(df_ses))
        for i_idx in range(len(df_ses)):
            choice = choices[i_idx]
            if choice == "I":  # no models track ignore now
                chosen_values[i_idx] = np.nan
                chosen_kernels[i_idx] = np.nan
                chosen_probabilities[i_idx] = np.nan
                chosen_stay_probabilities[i_idx] = np.nan
                # chosen_licks[i_idx] = np.nan
            else:
                chosen_values[i_idx] = df_ses["Q_" + {"L": "left", "R": "right"}[choice]].values[
                    i_idx
                ]
                if has_kernel:
                    chosen_kernels[i_idx] = df_ses[choice + "_kernel"].values[i_idx]
                chosen_probabilities[i_idx] = df_ses[choice + "_prob"].values[i_idx]
                if choice != "I":
                    unchosen_values[i_idx] = df_ses[
                        "Q_" + {"L": "right", "R": "left"}[choice]
                    ].values[
                        i_idx
                    ]  # noqa: E501
                    unchosen_probabilities[i_idx] = df_ses[
                        {"L": "R", "R": "L"}[choice] + "_prob"
                    ].values[
                        i_idx
                    ]  # noqa: E501
                    if has_kernel:
                        unchosen_kernels[i_idx] = df_ses[
                            {"L": "R", "R": "L"}[choice] + "_kernel"
                        ].values[
                            i_idx
                        ]  # noqa: E501
                    # chosen_licks[i_idx] = df_ses['licks '+choice].values[i_idx]
                if i_idx < len(df_ses) - 1:
                    chosen_stay_probabilities[i_idx] = df_ses[choice + "_prob"].values[i_idx + 1]
        for i_mod, mod in enumerate(models):
            df_ses.loc[:, "Q_chosen"] = chosen_values
            df_ses.loc[:, "Q_unchosen"] = unchosen_values
            df_ses.loc[:, "Q_sum"] = df_ses["Q_left"].values + df_ses["Q_right"].values
            df_ses.loc[:, "Q_Delta"] = df_ses["Q_chosen"].values - df_ses["Q_unchosen"].values
            df_ses.loc[:, "Q_change"] = np.concatenate([[0], np.diff(chosen_values)])

            df_ses.loc[:, "P_chosen"] = chosen_probabilities
            df_ses.loc[:, "P_unchosen"] = unchosen_probabilities
            df_ses.loc[:, "P_sum"] = df_ses["L_prob"].values + df_ses["R_prob"].values
            df_ses.loc[:, "P_Delta"] = df_ses["P_chosen"].values - df_ses["P_unchosen"].values
            df_ses.loc[:, "P_change"] = np.concatenate([[0], np.diff(chosen_probabilities)])

            if has_kernel:
                df_ses.loc[:, "K_chosen"] = chosen_kernels
                df_ses.loc[:, "K_unchosen"] = unchosen_kernels
                df_ses.loc[:, "K_sum"] = df_ses["L_kernel"].values + df_ses["R_kernel"].values
                df_ses.loc[:, "K_Delta"] = df_ses["K_chosen"].values - df_ses["K_unchosen"].values
                df_ses.loc[:, "K_change"] = np.concatenate([[0], np.diff(chosen_kernels)])

            df_ses.loc[:, "Cprobstay"] = chosen_stay_probabilities
            df_ses.loc[:, "RPE_earned"] = df_ses["earned_reward"].astype(float) - chosen_values
            df_ses.loc[:, "RPE_all"] = (
                df_ses["earned_reward"].astype(float) + df_ses["extra_reward"].astype(float)
            ) - chosen_values
            df_ses.loc[pd.isna(df_ses["choice"]), "RPE_earned"] = np.nan
            df_ses.loc[pd.isna(df_ses["choice"]), "RPE_all"] = np.nan
            # df_ses.loc[:,['licks_chosen']] = chosen_licks
        df_trials_fm_enriched = pd.concat([df_trials_fm_enriched, df_ses], axis=0)
    return df_trials_fm_enriched
